/*
!===============================================================================
!
! cublas_wrap.c - cuda version for some lapack functions
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 2011 Yoshifumi Nakamura
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
*/
#ifdef NamesToLower
# define  ZMINV_NV  zminv_nv
# define ZGETF2_NV zgetf2_nv
# define  ZGEMM_NV  zgemm_nv
#endif
#ifdef NamesToLower_
# define  ZMINV_NV  zminv_nv_
# define ZGETF2_NV zgetf2_nv_
# define  ZGEMM_NV  zgemm_nv_
#endif
#ifdef NamesToLower__
# define  ZMINV_NV  zminv_nv__
# define ZGETF2_NV zgetf2_nv__
# define  ZGEMM_NV  zgemm_nv__
#endif

#include <cublas.h>
#include <stdio.h>
#define min(a, b) ((a) < (b) ? (a) : (b))

void  ZMINV_NV(const int* m, const int* n, const double* A, const int* lda, 
	       int* ipiv, int* info){

  //  printf ("now zminv_nv_ started\n");

  int stat, i;
  cuDoubleComplex *devA, *devW, *zero;
  cuDoubleComplex mone, one, tmp, ttmp;
   one=make_cuDoubleComplex( 1, 0);
  mone=make_cuDoubleComplex(-1, 0);

  zero=( cuDoubleComplex * )malloc(*n *16);
  for (i=0; i< *n; i++){ zero[i]=make_cuDoubleComplex( 0, 0);}


  *info=0;
  stat = cublasInit();
  if (stat != CUBLAS_STATUS_SUCCESS) {
    printf ("cublasInit failed");
    *info =1;
    return;
  }

  int mm=(*m);
  int nn=(*n);
  int size=mm * nn;

  stat = cublasAlloc (size, 16, (void**)&devA);
  if (stat != CUBLAS_STATUS_SUCCESS) {
    printf ("device memory allocation failed %d %d\n",stat,size);
    cublasShutdown();
    *info=1;
    return;
  }
  stat = cublasAlloc (nn, 16, (void**)&devW);
  if (stat != CUBLAS_STATUS_SUCCESS) {
    printf ("device memory allocation failed %d %d\n",stat,nn);
    cublasShutdown();
    *info=1;
    return;
  }
  cublasSetMatrix(*m, *n, 16, A, *n, devA, *m);

//ZGETF2 start
  int j, j0n, j1n, jp;
  for (j=0; j<min(mm,nn); j++) {
    j0n= j     * nn;
    j1n= (j+1) * nn;
    jp = j - 1 + cublasIzamax( mm-j, &devA[ j+j0n ], 1 ) ;

    ipiv[ j ] = jp + 1 ;
    //    if (j0n+j  >= nn*mm || j0n+j  <0){printf ("zgetf2_nv_ j0n+j  =%d \n",j0n+j  );exit;}
    //    if (j0n+jp >= nn*mm || j0n+jp <0){printf ("zgetf2_nv_ j0n+jp =%d \n",j0n+jp );exit;}
    //    if (j      >= nn*mm || j      <0){printf ("zgetf2_nv_ j      =%d \n",j      );exit;}
    //    if (jp     >= nn*mm || jp     <0){printf ("zgetf2_nv_ jp     =%d \n",jp     );exit;}
    
    cublasGetMatrix(1, 1, 16,  &devA[ j0n+jp ], 1, (void*)&tmp, 1);

    if(!( tmp.x == 0.0d && tmp.y == 0.0d) ) {
      if( jp != j) cublasZswap(nn,  &devA[ j ], *lda, &devA[ jp ], *lda);
      cublasGetMatrix(1, 1, 16,  &devA[j0n+j], 1, (void*)&tmp, 1);  
      ttmp = cuCdiv(one , tmp);
      if( j  < mm-1) cublasZscal(mm-j-1, ttmp, &devA[ j+1+j0n], 1);
    } else if( *info == 0 ) {
      printf ("zgetf2_nv_ tmp=0 at j=%d\n",j);  
      *info = j;
    }
    if ( j < min(mm,nn)-1)
      cublasZgeru( mm-j-1, nn-j-1, mone, &devA[ j+1+j0n ], 1, &devA[ j+j1n ], *lda, &devA[ j+1+j1n ], *lda);
    //ZGERU( M-J, N-J, -ONE, A( J+1, J ), 1, A( J, J+1 ), LDA, A( J+1, J+1 ), LDA )
  }
//ZGETF2 end

//ZTRTI2 for Upper & No-unit start
  for (j=0;j<nn;j++) {
    j0n= j     * nn;
    cublasGetMatrix(1, 1, 16,  &devA[j0n+j], 1, (void*)&tmp, 1);
    ttmp = cuCdiv(one , tmp);
    cublasSetMatrix(1, 1, 16,  (void*)&ttmp, 1, &devA[j0n+j], 1);
    ttmp.x = -ttmp.x;
    ttmp.y = -ttmp.y;
    cublasZtrmv('U','N', 'N', j, devA, *lda, &devA[j0n],1);
    cublasZscal( j, ttmp, &devA[j0n],1);
  }
//ZTRTI2 for Upper & No-unit end

//===============================================================
//ZGETRI unblocked start
  for (j=nn-1;j>=0;j--) {
    j0n= j     * nn;
    j1n= (j+1) * nn;
    if (j!=nn-1)cublasZcopy(nn-j-1, &devA[j0n+j+1], 1, devW, 1);
    if (j!=nn-1)cublasSetVector(nn-j-1,   16, zero, 1, &devA[j0n+j+1], 1);
    if( j < nn-1 )
      cublasZgemv('N', nn, nn-j-1, mone, &devA[j1n], *lda, devW, 1, one, &devA[ j0n],1);   
  }
//ZGETRI unblocked end

// interchange
  for (j=nn-2;j>=0;j--) {
    jp=ipiv[j]-1;
    if (jp != j) cublasZswap(nn, &devA[j*nn],1, &devA[jp*nn],1);
  }

  free(zero);
  cublasGetMatrix(*m, *n, 16, devA, *n, (void*)A, *m);
  stat = cublasFree(devA);
  if (stat != CUBLAS_STATUS_SUCCESS) {
    printf ("cublasFree failed");
    *info =1;
    return;
  }

  stat = cublasFree(devW);
  if (stat != CUBLAS_STATUS_SUCCESS) {
    printf ("cublasFree failed");
    *info =1;
    return;
  }

  cublasShutdown();
  //if (stat != CUBLAS_STATUS_SUCCESS) {
  //  printf ("cublasShutdown failed");
  //  *info =1;
  //  return;
  // }
  //fflush(stdout);
}


void ZGETF2_NV(const int* m, const int* n, const double* A, const int* lda, 
	       int* ipiv, int* info){
  int stat, i;
  cuDoubleComplex *devA, *devW, *zero;
  cuDoubleComplex mone, one, tmp, ttmp;
   one=make_cuDoubleComplex( 1, 0);
  mone=make_cuDoubleComplex(-1, 0);

  zero=( cuDoubleComplex * )malloc(*n *16);
  for (i=0; i< *n; i++){ zero[i]=make_cuDoubleComplex( 0, 0);}


  *info=0;
  stat = cublasInit();
  if (stat != CUBLAS_STATUS_SUCCESS) {
    printf ("cublasInit failed");
    *info =1;
    return;
  }

  int mm=(*m);
  int nn=(*n);
  int size=mm * nn;

  stat = cublasAlloc (size, 16, (void**)&devA);
  if (stat != CUBLAS_STATUS_SUCCESS) {
    printf ("device memory allocation failed %d %d\n",stat,size);
    cublasShutdown();
    *info=1;
    return;
  }
  stat = cublasAlloc (nn, 16, (void**)&devW);
  if (stat != CUBLAS_STATUS_SUCCESS) {
    printf ("device memory allocation failed %d %d\n",stat,nn);
    cublasShutdown();
    *info=1;
    return;
  }
  cublasSetMatrix(*m, *n, 16, A, *n, devA, *m);

//ZGETF2 start
  int j, j0n, j1n, jp;
  for (j=0; j<min(mm,nn); j++) {
    j0n= j     * nn;
    j1n= (j+1) * nn;
    jp = j - 1 + cublasIzamax( mm-j, &devA[ j+j0n ], 1 ) ;

    ipiv[ j ] = jp + 1 ;
    //    if (j0n+j  >= nn*mm || j0n+j  <0){printf ("zgetf2_nv_ j0n+j  =%d \n",j0n+j  );exit;}
    //    if (j0n+jp >= nn*mm || j0n+jp <0){printf ("zgetf2_nv_ j0n+jp =%d \n",j0n+jp );exit;}
    //    if (j      >= nn*mm || j      <0){printf ("zgetf2_nv_ j      =%d \n",j      );exit;}
    //    if (jp     >= nn*mm || jp     <0){printf ("zgetf2_nv_ jp     =%d \n",jp     );exit;}
    
    cublasGetMatrix(1, 1, 16,  &devA[ j0n+jp ], 1, (void*)&tmp, 1);

    if(!( tmp.x == 0.0d && tmp.y == 0.0d) ) {
      if( jp != j) cublasZswap(nn,  &devA[ j ], *lda, &devA[ jp ], *lda);
      cublasGetMatrix(1, 1, 16,  &devA[j0n+j], 1, (void*)&tmp, 1);  
      ttmp = cuCdiv(one , tmp);
      if( j  < mm-1) cublasZscal(mm-j-1, ttmp, &devA[ j+1+j0n], 1);
    } else if( *info == 0 ) {
      printf ("zgetf2_nv_ tmp=0 at j=%d\n",j);  
      *info = j;
    }
    if ( j < min(mm,nn)-1)
      cublasZgeru( mm-j-1, nn-j-1, mone, &devA[ j+1+j0n ], 1, &devA[ j+j1n ], *lda, &devA[ j+1+j1n ], *lda);
    //ZGERU( M-J, N-J, -ONE, A( J+1, J ), 1, A( J, J+1 ), LDA, A( J+1, J+1 ), LDA )
  }
//ZGETF2 end

  free(zero);
  cublasGetMatrix(*m, *n, 16, devA, *n, (void*)A, *m);
  stat = cublasFree(devA);
  if (stat != CUBLAS_STATUS_SUCCESS) {
    printf ("cublasFree failed");
    *info =1;
    return;
  }

  stat = cublasFree(devW);
  if (stat != CUBLAS_STATUS_SUCCESS) {
    printf ("cublasFree failed");
    *info =1;
    return;
  }
  cublasShutdown();
}


//================================================================================
    void ZGEMM_NV(const char *transa, const char *transb, const int *m,
		  const int *n, const int *k, const cuDoubleComplex  *alpha,
		  const cuDoubleComplex  *A, const int *lda,
		  const cuDoubleComplex  *B, const int *ldb, 
		  const cuDoubleComplex  *beta, const cuDoubleComplex *C,
		  const int *ldc, int *info)

{
  //  printf ("now zgemm_nv_ started\n");

  cuDoubleComplex *devA, *devB, *devC ;
  int size=(int)(*m) * (int)(*n) ;
  int stat;

  *info=0;
  cublasInit();
  //if (stat != CUBLAS_STATUS_SUCCESS) {
  //  printf ("cublasInit failed");
  //  *info =1;
  //  return;
  //}

  stat = cublasAlloc (size, 16, (void**)&devA);
  if (stat != CUBLAS_STATUS_SUCCESS) {
    printf ("device memory allocation failed %d %d\n",stat,size);
    *info =1;
    return;
  }
  stat = cublasAlloc (size, 16, (void**)&devB);
  if (stat != CUBLAS_STATUS_SUCCESS) {
    printf ("device memory allocation failed %d %d\n",stat,size);
    *info =1;
    return;
  }
  stat = cublasAlloc (size, 16, (void**)&devC);
  if (stat != CUBLAS_STATUS_SUCCESS) {
    printf ("device memory allocation failed %d %d\n",stat,size);
    *info =1;
    return;
  }
  //  printf ("cublasAlloc done\n");

  cublasSetMatrix(*m, *n, 16, A, *n, devA, *m);
  cublasSetMatrix(*m, *n, 16, B, *n, devB, *m);

  // cublasSetMatrix(*m, *n, 16, C, *n, devC, *m);
  //  printf ("cublasSetMatrix done\n");

  cublasZgemm (transa[0], transb[0], *m, *n, *k, *alpha, devA, *lda, devB, *ldb, *beta, devC, *ldc);
  //  cublasZgemm (transa[0], transb[0], *m, *n, *k, *alpha, devA, *lda, devB, *ldb, *beta, devC, *ldc);

  cublasGetMatrix(*m, *n, 16, devC, *n, (void*)C, *m);
  //   printf ("cublasGetMatrix done\n");

  stat = cublasFree(devA);
  if (stat != CUBLAS_STATUS_SUCCESS) {
    printf ("cublasFree failed");
    *info =1;
    return;
  }
  stat = cublasFree(devB);
  if (stat != CUBLAS_STATUS_SUCCESS) {
    printf ("cublasFree failed");
    *info =1;
    return;
  }
  stat = cublasFree(devC);
  if (stat != CUBLAS_STATUS_SUCCESS) {
    printf ("cublasFree failed");
    *info =1;
    return;
  }

  cublasShutdown();
  //if (stat != CUBLAS_STATUS_SUCCESS) {
  //  printf ("cublasShutdown failed");
  //  *info =1;
  //  return;
  //}
}
