!===============================================================================
!
! dirac_matrix.F90
!
! make Dirac operator as matrix form,
! matrix shape for parallel computing is (12, volume/NPE, 12, volume)
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 2011 Yoshifumi Nakamura
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# include "defs.h"

!-------------------------------------------------------------------------------
module module_dirac_matrix
  implicit none
  COMPLEX,dimension(4,4),save :: proj_1_p_gamma1, proj_1_m_gamma1
  COMPLEX,dimension(4,4),save :: proj_1_p_gamma2, proj_1_m_gamma2
  COMPLEX,dimension(4,4),save :: proj_1_p_gamma3, proj_1_m_gamma3
  COMPLEX,dimension(4,4),save :: proj_1_p_gammac, proj_1_m_gammac
  COMPLEX,dimension(4,4),save :: proj_1_p_gamman, proj_1_m_gamman

  integer, pointer, save :: ieo2_around_i(:,:,:)
  integer, pointer, save :: ieo2_around_is(:,:,:)
  integer, pointer, save :: ieo2_around_it(:,:,:)
end module

!-------------------------------------------------------------------------------
subroutine checkmat(id)
  use module_field
  use module_action
  use module_vol
  implicit none
  integer, intent(in) :: id
  SPINCOL_FIELD :: eta_e, eta_o, zeta_e, zeta_o
  COMPLEX,allocatable :: zeta(:), eta(:), zeta1(:), mat(:,:)
  REAL :: fac1, fac2

  call stderr2("Now check mat-vec mult with chemical potential")

  allocate(zeta(12*vol), zeta1(12*vol), eta(12*vol), mat(12*vol, 12*volume))

  call make_source(eta_e, eta_o, RANDOM_SOURCE)
  call sceo2sc(eta, eta_e, eta_o)
  !
  ! result multiplied by using d
  !
!  call d(se, so, zeta_e, eta_o, gauge(2)%u(1,1,1,0,1))
!  call sc_xpby(zeta_e, eta_e, - action%mtilde(id)%kappa)
!  call d(so, se, zeta_o, eta_e, gauge(2)%u(1,1,1,0,1))
!  call sc_xpby(zeta_o, eta_o, - action%mtilde(id)%kappa)
!  call clover_mult_a(zeta_e, clover(action%mtilde(id)%cid)%a(1,1,se), eta_e, volh)
!  call clover_mult_a(zeta_o, clover(action%mtilde(id)%cid)%a(1,1,so), eta_o, volh)

  fac1=exp( chemi)
  fac2=exp(-chemi)
  if (chemi /= 0)call emuu4(gauge(2)%u, fac1, fac2, .false.)
  call unprec_wmul(zeta_e, zeta_o, eta_e, eta_o, id)
  if (chemi /= 0)call emuu4(gauge(2)%u, ONE/fac1, ONE/fac2, .false.)

  call sceo2sc(zeta, zeta_e, zeta_o)
  call make_dirac_op(mat, id, exp(chemi), exp(-chemi))
  call matvec_mul(mat, zeta1, eta, 12*vol, 12*volume)
  call compare_sc(zeta, zeta1, 24*vol)

  deallocate(zeta, zeta1, eta, mat)
!!  call comm_finalize()
!!  call die("stop")
end

!-------------------------------------------------------------------------------
# define CLOVER_AS_COMPLEX_ARRAY
# include "clover.h"
subroutine clover_put(mat, eo, a, offset)
  use module_dirac_matrix
  use module_vol
  implicit none
  COMPLEX, dimension(18, 2, *) :: a
  COMPLEX:: mat(3,4,vol,3,4,volume), cl1(6,6), cl2(6,6), clp(6,6), clm(6,6)
  integer :: c1, s1, c2, s2, ii, i, eo, sc1, sc2, offset, jj

  call init_dirac_matrix()
  do i=1, volh
     ii=ieo2_around_i(i,eo,0)
     jj=ii + offset
# undef J
# define J 1
     cl1(1,1) =A11 ; cl1(1,2) =A12 ; cl1(1,3) =A13 ; cl1(1,4) =A14 ; cl1(1,5) =A15 ; cl1(1,6) =A16
     cl1(2,1) =A21 ; cl1(2,2) =A22 ; cl1(2,3) =A23 ; cl1(2,4) =A24 ; cl1(2,5) =A25 ; cl1(2,6) =A26
     cl1(3,1) =A31 ; cl1(3,2) =A32 ; cl1(3,3) =A33 ; cl1(3,4) =A34 ; cl1(3,5) =A35 ; cl1(3,6) =A36
     cl1(4,1) =A41 ; cl1(4,2) =A42 ; cl1(4,3) =A43 ; cl1(4,4) =A44 ; cl1(4,5) =A45 ; cl1(4,6) =A46
     cl1(5,1) =A51 ; cl1(5,2) =A52 ; cl1(5,3) =A53 ; cl1(5,4) =A54 ; cl1(5,5) =A55 ; cl1(5,6) =A56
     cl1(6,1) =A61 ; cl1(6,2) =A62 ; cl1(6,3) =A63 ; cl1(6,4) =A64 ; cl1(6,5) =A65 ; cl1(6,6) =A66
# undef J
# define J 2
     cl2(1,1) =A11 ; cl2(1,2) =A12 ; cl2(1,3) =A13 ; cl2(1,4) =A14 ; cl2(1,5) =A15 ; cl2(1,6) =A16
     cl2(2,1) =A21 ; cl2(2,2) =A22 ; cl2(2,3) =A23 ; cl2(2,4) =A24 ; cl2(2,5) =A25 ; cl2(2,6) =A26
     cl2(3,1) =A31 ; cl2(3,2) =A32 ; cl2(3,3) =A33 ; cl2(3,4) =A34 ; cl2(3,5) =A35 ; cl2(3,6) =A36
     cl2(4,1) =A41 ; cl2(4,2) =A42 ; cl2(4,3) =A43 ; cl2(4,4) =A44 ; cl2(4,5) =A45 ; cl2(4,6) =A46
     cl2(5,1) =A51 ; cl2(5,2) =A52 ; cl2(5,3) =A53 ; cl2(5,4) =A54 ; cl2(5,5) =A55 ; cl2(5,6) =A56
     cl2(6,1) =A61 ; cl2(6,2) =A62 ; cl2(6,3) =A63 ; cl2(6,4) =A64 ; cl2(6,5) =A65 ; cl2(6,6) =A66

     clp=cl1+cl2
     clm=cl1-cl2
     do s1=1,2;   do s2=1,2
     do c1=1,3;   do c2=1,3
        sc1=(s1-1)*3+c1
        sc2=(s2-1)*3+c2
        mat(c1,s1  ,ii,c2,s2  ,jj)=clp(sc1,sc2)
        mat(c1,s1+2,ii,c2,s2+2,jj)=clp(sc1,sc2)
        mat(c1,s1  ,ii,c2,s2+2,jj)=clm(sc1,sc2)
        mat(c1,s1+2,ii,c2,s2  ,jj)=clm(sc1,sc2)
     enddo; enddo
     enddo; enddo
  enddo

end

!-------------------------------------------------------------------------------
subroutine clover_put_3d(mat, eo, a, offset)
  use module_dirac_matrix
  use module_vol
  use module_lattice
  implicit none
  COMPLEX, dimension(18, 2, *) :: a
  COMPLEX:: mat(3,4,nx*ny*nz,3,4,lx*ly*lz,lt), cl1(6,6), cl2(6,6), clp(6,6), clm(6,6)
  integer :: c1, s1, c2, s2, is, it, i, eo, sc1, sc2, jj, offset

  call init_dirac_matrix()
  do i=1, volh
     is=ieo2_around_is(i,eo,0)
     it=ieo2_around_it(i,eo,0)
     jj=is + offset
# undef J
# define J 1
     cl1(1,1) =A11 ; cl1(1,2) =A12 ; cl1(1,3) =A13 ; cl1(1,4) =A14 ; cl1(1,5) =A15 ; cl1(1,6) =A16
     cl1(2,1) =A21 ; cl1(2,2) =A22 ; cl1(2,3) =A23 ; cl1(2,4) =A24 ; cl1(2,5) =A25 ; cl1(2,6) =A26
     cl1(3,1) =A31 ; cl1(3,2) =A32 ; cl1(3,3) =A33 ; cl1(3,4) =A34 ; cl1(3,5) =A35 ; cl1(3,6) =A36
     cl1(4,1) =A41 ; cl1(4,2) =A42 ; cl1(4,3) =A43 ; cl1(4,4) =A44 ; cl1(4,5) =A45 ; cl1(4,6) =A46
     cl1(5,1) =A51 ; cl1(5,2) =A52 ; cl1(5,3) =A53 ; cl1(5,4) =A54 ; cl1(5,5) =A55 ; cl1(5,6) =A56
     cl1(6,1) =A61 ; cl1(6,2) =A62 ; cl1(6,3) =A63 ; cl1(6,4) =A64 ; cl1(6,5) =A65 ; cl1(6,6) =A66
# undef J
# define J 2
     cl2(1,1) =A11 ; cl2(1,2) =A12 ; cl2(1,3) =A13 ; cl2(1,4) =A14 ; cl2(1,5) =A15 ; cl2(1,6) =A16
     cl2(2,1) =A21 ; cl2(2,2) =A22 ; cl2(2,3) =A23 ; cl2(2,4) =A24 ; cl2(2,5) =A25 ; cl2(2,6) =A26
     cl2(3,1) =A31 ; cl2(3,2) =A32 ; cl2(3,3) =A33 ; cl2(3,4) =A34 ; cl2(3,5) =A35 ; cl2(3,6) =A36
     cl2(4,1) =A41 ; cl2(4,2) =A42 ; cl2(4,3) =A43 ; cl2(4,4) =A44 ; cl2(4,5) =A45 ; cl2(4,6) =A46
     cl2(5,1) =A51 ; cl2(5,2) =A52 ; cl2(5,3) =A53 ; cl2(5,4) =A54 ; cl2(5,5) =A55 ; cl2(5,6) =A56
     cl2(6,1) =A61 ; cl2(6,2) =A62 ; cl2(6,3) =A63 ; cl2(6,4) =A64 ; cl2(6,5) =A65 ; cl2(6,6) =A66

     clp=cl1+cl2
     clm=cl1-cl2
     do s1=1,2;   do s2=1,2
     do c1=1,3;   do c2=1,3
        sc1=(s1-1)*3+c1
        sc2=(s2-1)*3+c2
        mat(c1,s1  ,is,c2,s2  ,jj,it)=clp(sc1,sc2)
        mat(c1,s1+2,is,c2,s2+2,jj,it)=clp(sc1,sc2)
        mat(c1,s1  ,is,c2,s2+2,jj,it)=clm(sc1,sc2)
        mat(c1,s1+2,is,c2,s2  ,jj,it)=clm(sc1,sc2)
     enddo; enddo
     enddo; enddo
  enddo

end

!-------------------------------------------------------------------------------
subroutine one_put(mat, m, n, offset)
  implicit none
  integer, intent(in) :: m, n, offset
  COMPLEX,intent(inout) :: mat(12,m,12,n)
  integer :: i

  do i = 1, m
     mat( 1, i, 1, i + offset )=ONE
     mat( 2, i, 2, i + offset )=ONE
     mat( 3, i, 3, i + offset )=ONE
     mat( 4, i, 4, i + offset )=ONE
     mat( 5, i, 5, i + offset )=ONE
     mat( 6, i, 6, i + offset )=ONE
     mat( 7, i, 7, i + offset )=ONE
     mat( 8, i, 8, i + offset )=ONE
     mat( 9, i, 9, i + offset )=ONE
     mat(10, i,10, i + offset )=ONE
     mat(11, i,11, i + offset )=ONE
     mat(12, i,12, i + offset )=ONE
  enddo
end

!-------------------------------------------------------------------------------
subroutine one_put_3d(mat, m, n, offset, t)
  implicit none
  integer, intent(in) :: m, n, offset, t
  COMPLEX,intent(inout) :: mat(12, m, 12, n, t)
  integer :: i

  do i = 1, t
     call one_put(mat(:,:,:,:,i), m, n, offset)
  enddo
end

!-------------------------------------------------------------------------------
subroutine hop_put(mat, gp, gm, i, eo, dir, ii, kappa, fac1, fac2)
  use module_nn
  use module_vol
  use module_dirac_matrix
  use module_field
  implicit none
  integer, intent(in) :: i, eo, dir, ii
  COMPLEX, intent(in) :: gp(4,4), gm(4,4)
  COMPLEX, intent(inout) :: mat(3,4,vol,3,4,volume)
  REAL, intent(in) :: kappa, fac1, fac2
  integer :: s1, s2, col1, col2, jm, jj, jb

  jm=nn(i, eo, dir, BWD)
  jj=ieo2_around_i(i,eo,1 + 2*(dir-1) )
  jb=ieo2_around_i(i,eo,2 + 2*(dir-1) )

  do s1 = 1, 4
  do s2 = 1, 4
     if ( gp(s1, s2) /=0 ) then
        do col1 = 1, 3
        do col2 = 1, 3
           mat(col1,s1,ii,col2,s2,jj)=mat(col1,s1,ii,col2,s2,jj) &
          -kappa*      gauge(2)%u(col1,col2, i,   eo,dir) *gp(s1,s2)*fac1
        enddo
        enddo
     endif
     if ( gm(s1, s2) /=0 ) then
        do col1 = 1, 3
        do col2 = 1, 3
           mat(col1,s1,ii,col2,s2,jb)=mat(col1,s1,ii,col2,s2,jb) &
          -kappa*conjg(gauge(2)%u(col2,col1, jm,1-eo,dir))*gm(s1,s2)*fac2
        enddo
        enddo
     endif
  enddo
  enddo
end

!-------------------------------------------------------------------------------
subroutine hop_put_3d(mat, gp, gm, i, eo, dir, is, it, itoffset, kappa)
  use module_nn
  use module_lattice
  use module_dirac_matrix
  use module_field
  implicit none
  integer, intent(in) :: i, eo, dir, is, it, itoffset
  COMPLEX, intent(in) :: gp(4,4), gm(4,4)
  COMPLEX, intent(inout) :: mat(3,4,nx*ny*nz,3,4,lx*ly*lz,lt)
  REAL, intent(in) :: kappa
  integer :: s1, s2, col1, col2, jm, jj, jb, it1

  jm=nn(i, eo, dir, BWD)
  jj=ieo2_around_is(i,eo,1 + 2*(dir-1) )
  jb=ieo2_around_is(i,eo,2 + 2*(dir-1) )
  it1=it + itoffset

  do s1 = 1, 4
  do s2 = 1, 4
     if ( gp(s1, s2) /=0 ) then
        do col1 = 1, 3
        do col2 = 1, 3
           mat(col1,s1,is,col2,s2,jj, it)=mat(col1,s1,is,col2,s2,jj,it) &
          -kappa*      gauge(2)%u(col1,col2, i,   eo,dir) *gp(s1,s2)
        enddo
        enddo
     endif
     if ( gm(s1, s2) /=0 ) then
        do col1 = 1, 3
        do col2 = 1, 3
           mat(col1,s1,is,col2,s2,jb, it1)=mat(col1,s1,is,col2,s2,jb,it1) &
          -kappa*conjg(gauge(2)%u(col2,col1, jm,1-eo,dir))*gm(s1,s2)
        enddo
        enddo
     endif
  enddo
  enddo
end

!-------------------------------------------------------------------------------
subroutine make_dirac_op(mat, id, fac1, fac2) ! for one node
  use module_dirac_matrix
  use module_vol
  use module_field
  use module_action
  use module_function_decl
  implicit none
  integer, intent(in) :: id
  REAL, intent(in) :: fac1, fac2
  COMPLEX,intent(out) :: mat(3,4,vol,3,4,volume)
  REAL :: kappa
  integer :: ii, i, eo, offset

  call init_dirac_matrix()
  kappa=action%mtilde(id)%kappa

  mat=ZERO
  offset =  my_pe()*vol
  if (action%mtilde(id)%cswkappa == 0) then
     call one_put(mat, vol, volume, offset)
  else
     call clover_put(mat, EVEN, clover(action%mtilde(id)%cid)%a(1,1,se), offset)
     call clover_put(mat, ODD,  clover(action%mtilde(id)%cid)%a(1,1,so), offset)
  endif

  if (kappa /= 0) then
     do eo=EVEN, ODD
     do i=1, volh
        ii=ieo2_around_i(i,eo,0)
        call hop_put(mat, proj_1_p_gamma1, proj_1_m_gamma1, i, eo, 1, ii, kappa, ONE, ONE) ! x-dir
        call hop_put(mat, proj_1_p_gamma2, proj_1_m_gamma2, i, eo, 2, ii, kappa, ONE, ONE) ! y-dir
        call hop_put(mat, proj_1_p_gamma3, proj_1_m_gamma3, i, eo, 3, ii, kappa, ONE, ONE) ! z-dir
        call hop_put(mat, proj_1_m_gamman, proj_1_p_gamman, i, eo, 4, ii, kappa,fac1,fac2) ! t-dir
     enddo
     enddo
  endif

end

!-------------------------------------------------------------------------------
subroutine make_dirac_op_3d(mat3,matt, id) ! for one node
  use module_dirac_matrix
  use module_lattice
  use module_vol
  use module_nn
  use module_field
  use module_action
  use module_function_decl
  implicit none
  integer, intent(in) :: id
  REAL :: kappa
  COMPLEX:: mat3(3,4,nx*ny*nz,3,4,lx*ly*lz,lt)
  COMPLEX:: matt(3,4,nx*ny*nz,3,4,lx*ly*lz,2*lt)
  integer :: col1, s1, col2, s2, is, it, jj, i, j, eo, jb, jm, offset

  call init_dirac_matrix()
  kappa=action%mtilde(id)%kappa

  mat3=ZERO
  matt=ZERO
  offset =  my_pe()*vol/nt
  if (action%mtilde(id)%cswkappa == 0) then
     call one_put_3d(mat3, vol/nt, volume/nt, offset, nt)
!     do it=1,nt
!     do i=1,vol/nt
!     do col1=1,3
!     do s1=1,4
!        mat3(col1,s1,i, col1,s1,i,it)=ONE
!     enddo
!     enddo
!     enddo
!     enddo
  else
     call clover_put_3d(mat3, EVEN, clover(action%mtilde(id)%cid)%a(1,1,se), offset)
     call clover_put_3d(mat3, ODD,  clover(action%mtilde(id)%cid)%a(1,1,so), offset)
  endif

  if (kappa /= 0) then
     do eo=EVEN, ODD
     do i=1, volh
        is=ieo2_around_is(i,eo,0)
        it=ieo2_around_it(i,eo,0)
        call hop_put_3d(mat3, proj_1_p_gamma1, proj_1_m_gamma1, i, eo, 1, is, it, 0, kappa) ! x-dir
        call hop_put_3d(mat3, proj_1_p_gamma2, proj_1_m_gamma2, i, eo, 2, is, it, 0, kappa) ! y-dir
        call hop_put_3d(mat3, proj_1_p_gamma3, proj_1_m_gamma3, i, eo, 3, is, it, 0, kappa) ! z-dir
        call hop_put_3d(matt, proj_1_m_gamman, proj_1_p_gamman, i, eo, 4, is, it,lt, kappa) ! t-dir

!!        !
!!        ! x-dir
!!        !
!!        jj=ieo2_around_is(i,eo,1)
!!        jb=ieo2_around_is(i,eo,2)
!!        jm=nn(i, eo, 1, BWD)
!!        do s1=1,4;   do s2=1,4
!!        if (proj_1_p_gamma1(s1,s2) /=0) then
!!           do col1=1,3;   do col2=1,3
!!           mat3(col1,s1,is,col2,s2,jj,it)=mat3(col1,s1,is,col2,s2,jj,it) -kappa*      gauge(1)%u(col1,col2, i,   eo,1) *proj_1_p_gamma1(s1,s2)
!!           mat3(col1,s1,is,col2,s2,jb,it)=mat3(col1,s1,is,col2,s2,jb,it) -kappa*conjg(gauge(1)%u(col2,col1, jm,1-eo,1))*proj_1_m_gamma1(s1,s2)
!!           enddo;   enddo
!!        endif
!!        enddo;   enddo
!!        !
!!        ! y-dir
!!        !
!!        jj=ieo2_around_is(i,eo,3)
!!        jb=ieo2_around_is(i,eo,4)
!!        jm=nn(i, eo, 2, BWD)
!!        do s1=1,4;   do s2=1,4
!!        if (proj_1_p_gamma2(s1,s2) /=0) then
!!           do col1=1,3;   do col2=1,3
!!           mat3(col1,s1,is,col2,s2,jj,it)=mat3(col1,s1,is,col2,s2,jj,it) -kappa*      gauge(1)%u(col1,col2, i,   eo,2) *proj_1_p_gamma2(s1,s2)
!!           mat3(col1,s1,is,col2,s2,jb,it)=mat3(col1,s1,is,col2,s2,jb,it) -kappa*conjg(gauge(1)%u(col2,col1, jm,1-eo,2))*proj_1_m_gamma2(s1,s2)
!!           enddo;   enddo
!!        endif
!!        enddo;   enddo
!!        !
!!        ! z-dir
!!        !
!!        jj=ieo2_around_is(i,eo,5)
!!        jb=ieo2_around_is(i,eo,6)
!!        jm=nn(i, eo, 3, BWD)
!!        do s1=1,4;   do s2=1,4
!!        if (proj_1_p_gamma3(s1,s2) /=0) then
!!           do col1=1,3;   do col2=1,3
!!           mat3(col1,s1,is,col2,s2,jj,it)=mat3(col1,s1,is,col2,s2,jj,it) -kappa*      gauge(1)%u(col1,col2, i,   eo,3) *proj_1_p_gamma3(s1,s2)
!!           mat3(col1,s1,is,col2,s2,jb,it)=mat3(col1,s1,is,col2,s2,jb,it) -kappa*conjg(gauge(1)%u(col2,col1, jm,1-eo,3))*proj_1_m_gamma3(s1,s2)
!!           enddo;   enddo
!!        endif
!!        enddo;   enddo
!!        !
!!        ! t-dir
!!        !
!!        jj=ieo2_around_is(i,eo,7)
!!        jb=ieo2_around_is(i,eo,8)
!!        jm=nn(i, eo, 4, BWD)
!!        do s1=1,4;   do s2=1,4
!!        if (proj_1_m_gamman(s1,s2) /=0) then
!!           do col1=1,3;   do col2=1,3
!!           matt(col1,s1,is,col2,s2,jj,it)=matt(col1,s1,is,col2,s2,jj,it) -kappa*      gauge(2)%u(col1,col2, i,   eo,4) *proj_1_m_gamman(s1,s2)
!!           enddo;   enddo
!!        endif
!!        if (proj_1_p_gamman(s1,s2) /=0) then
!!           do col1=1,3;   do col2=1,3
!!           matt(col1,s1,is,col2,s2,jb,it+nt)=matt(col1,s1,is,col2,s2,jb,it+nt) -kappa*conjg(gauge(2)%u(col2,col1, jm,1-eo,4))*proj_1_p_gamman(s1,s2)
!!           enddo;   enddo
!!        endif
!!        enddo;   enddo
     enddo
     enddo
  endif

end

!-------------------------------------------------------------------------------
subroutine init_dirac_matrix()
  use module_dirac_matrix
  use module_lattice
  use module_vol
  implicit none
  integer :: i, eo, ll(4)

  if (.not. associated(ieo2_around_i)) then
  allocate(ieo2_around_i(volh,0:1, 0:8))
  allocate(ieo2_around_is(volh,0:1, 0:8))
  allocate(ieo2_around_it(volh,0:1, 0:8))
  proj_1_m_gamma1=0
  proj_1_p_gamma1=0
  proj_1_m_gamma2=0
  proj_1_p_gamma2=0
  proj_1_m_gamma3=0
  proj_1_p_gamma3=0
  proj_1_m_gammac=0
  proj_1_p_gammac=0
  proj_1_m_gamman=0
  proj_1_p_gamman=0

  do i=1,4
  proj_1_m_gamma1(i,i)=ONE
  proj_1_p_gamma1(i,i)=ONE
  proj_1_m_gamma2(i,i)=ONE
  proj_1_p_gamma2(i,i)=ONE
  proj_1_m_gamma3(i,i)=ONE
  proj_1_p_gamma3(i,i)=ONE
  proj_1_m_gammac(i,i)=ONE
  proj_1_p_gammac(i,i)=ONE
  enddo

  proj_1_m_gamman(3,3)=TWO
  proj_1_m_gamman(4,4)=TWO
  proj_1_p_gamman(1,1)=TWO
  proj_1_p_gamman(2,2)=TWO

  proj_1_m_gamma1(1,4)=cmplx(ZERO,ONE)
  proj_1_m_gamma1(2,3)=cmplx(ZERO,ONE)
  proj_1_m_gamma1(3,2)=cmplx(ZERO,-ONE)
  proj_1_m_gamma1(4,1)=cmplx(ZERO,-ONE)
  proj_1_p_gamma1(1,4)=cmplx(ZERO,-ONE)
  proj_1_p_gamma1(2,3)=cmplx(ZERO,-ONE)
  proj_1_p_gamma1(3,2)=cmplx(ZERO,ONE)
  proj_1_p_gamma1(4,1)=cmplx(ZERO,ONE)

  proj_1_m_gamma2(1,4)= ONE
  proj_1_m_gamma2(2,3)=-ONE
  proj_1_m_gamma2(3,2)=-ONE
  proj_1_m_gamma2(4,1)= ONE
  proj_1_p_gamma2(1,4)=-ONE
  proj_1_p_gamma2(2,3)= ONE
  proj_1_p_gamma2(3,2)= ONE
  proj_1_p_gamma2(4,1)=-ONE

  proj_1_m_gamma3(1,3)=cmplx(ZERO, ONE)
  proj_1_m_gamma3(2,4)=cmplx(ZERO,-ONE)
  proj_1_m_gamma3(3,1)=cmplx(ZERO,-ONE)
  proj_1_m_gamma3(4,2)=cmplx(ZERO, ONE)
  proj_1_p_gamma3(1,3)=cmplx(ZERO,-ONE)
  proj_1_p_gamma3(2,4)=cmplx(ZERO, ONE)
  proj_1_p_gamma3(3,1)=cmplx(ZERO, ONE)
  proj_1_p_gamma3(4,2)=cmplx(ZERO,-ONE)

  proj_1_m_gammac(1,3)=-ONE
  proj_1_m_gammac(2,4)=-ONE
  proj_1_m_gammac(3,1)= ONE
  proj_1_m_gammac(4,2)= ONE
  proj_1_p_gammac(1,3)= ONE
  proj_1_p_gammac(2,4)= ONE
  proj_1_p_gammac(3,1)=-ONE
  proj_1_p_gammac(4,2)=-ONE


  do eo=EVEN, ODD
  do i=1, volh
     call i2xyzt(i,     eo, ll)
     ieo2_around_i(i,eo,0)= 1 + ll(1) + ll(2)*nx + ll(3)*nx*ny + ll(4)*nx*ny*nz
     ieo2_around_is(i,eo,0)= 1 + ll(1) + ll(2)*nx + ll(3)*nx*ny
     ieo2_around_it(i,eo,0)= 1 + ll(4)

     call ieo_offset_global(i, eo, 1, FWD, ieo2_around_i(i,eo,1), ieo2_around_is(i,eo,1), ieo2_around_it(i,eo,1))
     call ieo_offset_global(i, eo, 1, BWD, ieo2_around_i(i,eo,2), ieo2_around_is(i,eo,2), ieo2_around_it(i,eo,2))
     call ieo_offset_global(i, eo, 2, FWD, ieo2_around_i(i,eo,3), ieo2_around_is(i,eo,3), ieo2_around_it(i,eo,3))
     call ieo_offset_global(i, eo, 2, BWD, ieo2_around_i(i,eo,4), ieo2_around_is(i,eo,4), ieo2_around_it(i,eo,4))
     call ieo_offset_global(i, eo, 3, FWD, ieo2_around_i(i,eo,5), ieo2_around_is(i,eo,5), ieo2_around_it(i,eo,5))
     call ieo_offset_global(i, eo, 3, BWD, ieo2_around_i(i,eo,6), ieo2_around_is(i,eo,6), ieo2_around_it(i,eo,6))
     call ieo_offset_global(i, eo, 4, FWD, ieo2_around_i(i,eo,7), ieo2_around_is(i,eo,7), ieo2_around_it(i,eo,7))
     call ieo_offset_global(i, eo, 4, BWD, ieo2_around_i(i,eo,8), ieo2_around_is(i,eo,8), ieo2_around_it(i,eo,8))
  enddo
  enddo

  endif


end

!-------------------------------------------------------------------------------
subroutine ieo_offset_global(i, eo , dir, fb, i4, i3, it)
  use module_function_decl
  use module_lattice
  use module_nnpe
  use module_nn
  use module_vol
  implicit none
  integer, intent(in) :: i, eo, dir, fb
  integer, intent(out):: i4, i3, it
  integer :: j, ll(4), pe(4), ip
  integer, save :: nxy, nxyz, count=0

  if (count==0) then
     nxy = nx * ny
     nxyz= nxy * nz
     count=-1
  endif

  pe = 0
  j = nn(i, eo, dir, fb)
  call i2xyzt(j, 1-eo, ll)

  if ( j > volh ) then
     if ( ll(dir) == -1 ) then
        pe(dir) = -1
        ll(dir) = n(dir)-1
     elseif( ll(dir) == n(dir) ) then
        pe(dir) = 1
        ll(dir) = 0
     endif
  endif

  ip=nnpe( pe(1), pe(2), pe(3), pe(4))

  it = 1 + ll(4)
  i3 = 1 + ll(1) + ll(2) * nx + ll(3) * nxy
  i4 = i3+ ll(4) * nxyz

  i3 = i3 + ip * nxyz
  i4 = i4 + ip * vol

end

!-------------------------------------------------------------------------------
subroutine compare_sc(in1, in2, size)
  use module_function_decl
  implicit none
  integer :: i, size
  REAL :: in1(size), in2(size), res, tmp

  res=0
  do i=1, size
     tmp=(in1(i)-in2(i))**2
     res=res+tmp
     if (tmp > 0.00001)write(*,*)i, tmp,in1(i),in2(i)
  enddo
  res=global_sum(res)
  call stderr2_real("vector difference",1,res)

end

!-------------------------------------------------------------------------------
subroutine sceo2sc(out, in_e, in_o)
  use module_lattice
  use module_vol
  implicit none
  integer :: id, ix, iy, iz, it, i, ii, j(4), spin, col

  SPINCOL_FIELD :: in_e, in_o
  COMPLEX       :: out(3,4,vol)
  integer, external :: std_xyzt2i, e_o

  i=0
  do it=0,nt-1
  do iz=0,nz-1
  do iy=0,ny-1
  do ix=0,nx-1
     i=i+1
     j = (/ix, iy, iz, it/)
     ii = std_xyzt2i(j)
     if (e_o(j) == EVEN) then
        do col=1, 3
        do spin=1, 4
           out(col,spin,i)=in_e(spin,col,ii)
        enddo
        enddo
     else
        do col=1, 3
        do spin=1, 4
           out(col,spin,i)=in_o(spin,col,ii)
        enddo
        enddo
     endif
  enddo
  enddo
  enddo
  enddo

end

!===============================================================================
! for checking
!-------------------------------------------------------------------------------
module phmc_read
  integer, save :: ipeo=0
 
  type su3_obj 
     complex(8) :: u(3,3) 
  end type

  type gsf_eo_wg_obj
     type(su3_obj) :: s(0:2,0:5,0:5,0:5)
     integer :: ieo,idummy(3)
  end type

  type gvf_eo_wg_obj
     type(gsf_eo_wg_obj)  :: mu(4)
     integer :: ieo,idummy(3)
  end type

  type gvf_wg_obj
     type(gvf_eo_wg_obj)  :: eo(0:1)
  end type

end

!-------------------------------------------------------------------------------
subroutine read_phmc_config()
  use phmc_read
  use module_vol
  use module_field
  implicit none
  type(gvf_wg_obj) :: phmc_u
  integer :: ix, iy, iz, it, ieo, imu, j(4), itb, jmu(4) , ee, jj, ic,jc
  integer :: ntx, nty, ntz, ntt
  integer :: ndx, ndy, ndz, ndt
  REAL :: plq, tmp
  GAUGE_FIELD ::u
  REAL, external :: sg
  integer, external :: std_xyzt2i, e_o

  write(*,*)"read_phmc_config: start"

  open(40,file="config.x0y0z0",status='unknown',form='unformatted')
  write(*,*)"read_phmc_config: open config.x0y0z0"

  read(40)phmc_u
  write(*,*)"read_phmc_config: have read config.x0y0z0"

  read(40)NTX,NTY,NTZ,NTT
  read(40)NDX,NDY,NDZ
  read(40)plq

  jmu(1)=4
  jmu(2)=3
  jmu(3)=2
  jmu(4)=1


  do imu = 1,4
  do it=1,4
     itb=it/2
  do iz=1,4
  do iy=1,4
  do ix=1,4
     ieo=mod(ix + iy + iz + it, 2)
     j=(/ ix-1 , iy-1, iz-1, it-1/)
     jj=std_xyzt2i(j)
     ee=e_o(j)
     u(:,:, jj ,ee, imu )=phmc_u%eo(ieo)%mu(imu)%s(itb,iz,iy,ix)%u
  enddo
  enddo
  enddo
  enddo
  enddo

  gauge(1)%u=u

  call xbound_g_field(gauge(1)%u)
  call d_g_init(gauge(1)%u)
  call conf_check(gauge(1)%u)

  !!call conf_hot(gauge(1)%u)
  tmp=sg(gauge(1)%u)
  write(*,*)"plaq value calculated by bqcd", ONE - tmp / (SIX * volume)

end

!===============================================================================
