!===============================================================================
!
! phmc_read.F90 - read phmc config
!
!-------------------------------------------------------------------------------
!
! Copyright (C) 2011 Yoshifumi Nakamura
!
! This file is part of BQCD -- Berlin Quantum ChromoDynamics program
!
! BQCD is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!
! BQCD is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with BQCD.  If not, see <http://www.gnu.org/licenses/>.
!
!-------------------------------------------------------------------------------
# include "defs.h"

!-------------------------------------------------------------------------------
module phmc_read
  integer, save :: ipeo=0
 
  type su3_obj 
     complex(8) :: u(3,3) 
  end type

  type gsf_eo_wg_obj
     type(su3_obj) :: s(0:2,0:5,0:5,0:5)
     integer :: ieo,idummy(3)
  end type

  type gvf_eo_wg_obj
     type(gsf_eo_wg_obj)  :: mu(4)
     integer :: ieo,idummy(3)
  end type

  type gvf_wg_obj
     type(gvf_eo_wg_obj)  :: eo(0:1)
  end type

end

!-------------------------------------------------------------------------------
subroutine read_phmc_config()
  use phmc_read
  use module_vol
  use module_field
  implicit none
  type(gvf_wg_obj) :: phmc_u
  integer :: ix, iy, iz, it, ieo, imu, j(4), itb, jmu(4) , ee, jj, ic,jc
  integer :: ntx, nty, ntz, ntt
  integer :: ndx, ndy, ndz, ndt
  REAL :: plq, tmp
  GAUGE_FIELD ::u
  REAL, external :: sg
  integer, external :: std_xyzt2i, e_o

  write(*,*)"read_phmc_config: start"

  open(40,file="config.x0y0z0",status='unknown',form='unformatted')
  write(*,*)"read_phmc_config: open config.x0y0z0"

  read(40)phmc_u
  write(*,*)"read_phmc_config: have read config.x0y0z0"

  read(40)NTX,NTY,NTZ,NTT
  read(40)NDX,NDY,NDZ
  read(40)plq

  jmu(1)=4
  jmu(2)=3
  jmu(3)=2
  jmu(4)=1

  do imu = 1,4
  do it=1,4
     itb=it/2
  do iz=1,4
  do iy=1,4
  do ix=1,4
     ieo=mod(ix + iy + iz + it, 2)
     j=(/ ix-1 , iy-1, iz-1, it-1/)
     jj=std_xyzt2i(j)
     ee=e_o(j)
     u(:,:, jj ,ee, imu )=phmc_u%eo(ieo)%mu(imu)%s(itb,iz,iy,ix)%u
  enddo
  enddo
  enddo
  enddo
  enddo

  gauge(1)%u=u

  call xbound_g_field(gauge(1)%u)
  call d_g_init(gauge(1)%u)
  call conf_check(gauge(1)%u)

  !!call conf_hot(gauge(1)%u)
  tmp=sg(gauge(1)%u)
  write(*,*)"plaq value calculated by bqcd", ONE - tmp / (SIX * volume)

end

!-------------------------------------------------------------------------------
subroutine read_phmc_sf_config()
  use module_vol
  use module_field
  implicit none
  integer :: ix, iy, iz, it, ieo, imu, j(4), itb, jmu(4) , ee, jj, ic,jc
  integer :: ntx, nty, ntz, ntt
  integer :: ndx, ndy, ndz, ndt
  COMPLEX, dimension(3,3,0:4,4,4,4,4) :: ue, uo
  REAL :: plq, tmp
  GAUGE_FIELD ::u
  REAL, external :: sg
  integer, external :: std_xyzt2i, e_o

  write(*,*)"read_phmc_sf_config: start"

  open(40,file="config.x0y0z0",status='unknown',form='unformatted')
  write(*,*)"read_phmc_config: open config.x0y0z0"

  read(40)ue, uo
  write(*,*)"read_phmc_config: have read config.x0y0z0"

  jmu(1)=4
  jmu(2)=3
  jmu(3)=2
  jmu(4)=1

  do imu = 1,4
  do it=1,8
     itb=it/2
  do iz=1,4
  do iy=1,4
  do ix=1,4
     ieo=mod(ix + iy + iz + it, 2)
     j=(/ ix-1 , iy-1, iz-1, it-1/)
     jj=std_xyzt2i(j)
     ee=e_o(j)
     if (ieo == 0) u(:,:, jj ,ee, imu ) = ue(:,:,itb,iz,iy,ix,imu)
     if (ieo == 1) u(:,:, jj ,ee, imu ) = uo(:,:,itb,iz,iy,ix,imu)
  enddo
  enddo
  enddo
  enddo
  enddo

  gauge(1)%u=u

  call xbound_g_field(gauge(1)%u)
  call d_g_init(gauge(1)%u)
  call conf_check(gauge(1)%u)

  !!call conf_hot(gauge(1)%u)
  tmp=sg(gauge(1)%u)
  write(*,*)"plaq value calculated by bqcd", ONE - tmp / (SIX * volume)
  call flush(6)
end

!===============================================================================
